@@ -12,6 +12,10 @@ import pytest
 
 from arroyo.backends.kafka import KafkaPayload
 from arroyo.processing.strategies import ProcessingStrategy
+from arroyo.processing.strategies.dead_letter_queue.policies.abstract import (
+    InvalidKafkaMessage,
+    InvalidMessages,
+)
 from arroyo.processing.strategies.streaming.collect import (
     CollectStep,
     ParallelCollectStep,
@@ -320,11 +324,11 @@ def test_message_batch() -> None:
     smm = SharedMemoryManager()
     smm.start()
 
-    block = smm.SharedMemory(4096)
-    assert block.size == 4096
+    block = smm.SharedMemory(16384)
+    assert block.size == 16384
 
     message = Message(
-        partition, 0, KafkaPayload(None, b"\x00" * 4000, []), datetime.now()
+        partition, 0, KafkaPayload(None, b"\x00" * 16000, []), datetime.now()
     )
 
     batch: MessageBatch[KafkaPayload] = MessageBatch(block)
@@ -356,13 +360,13 @@ def test_parallel_transform_worker_apply() -> None:
             KafkaPayload(None, b"\x00" * size, []),
             datetime.now(),
         )
-        for i, size in enumerate([1000, 1000, 2000, 4000])
+        for i, size in enumerate([4000, 4000, 8000, 12000])
     ]
 
     smm = SharedMemoryManager()
     smm.start()
-    input_block = smm.SharedMemory(8192)
-    assert input_block.size == 8192
+    input_block = smm.SharedMemory(32768)
+    assert input_block.size == 32768
 
     input_batch = MessageBatch[Any](input_block)
     for message in messages:
@@ -370,29 +374,29 @@ def test_parallel_transform_worker_apply() -> None:
 
     assert len(input_batch) == 4
 
-    output_block = smm.SharedMemory(4096)
-    assert output_block.size == 4096
+    output_block = smm.SharedMemory(16384)
+    assert output_block.size == 16384
 
-    index, output_batch = parallel_transform_worker_apply(
+    result = parallel_transform_worker_apply(
         transform_payload_expand,
         input_batch,
         output_block,
     )
 
     # The first batch should be able to fit 2 messages.
-    assert index == 2
-    assert len(output_batch) == 2
+    assert result.next_index_to_process == 2
+    assert len(result.valid_messages_transformed) == 2
 
-    index, output_batch = parallel_transform_worker_apply(
+    result = parallel_transform_worker_apply(
         transform_payload_expand,
         input_batch,
         output_block,
-        index,
+        result.next_index_to_process,
     )
 
     # The second batch should be able to fit one message.
-    assert index == 3
-    assert len(output_batch) == 1
+    assert result.next_index_to_process == 3
+    assert len(result.valid_messages_transformed) == 1
 
     # The last message is too large to fit in the batch.
     with pytest.raises(ValueTooLarge):
@@ -400,11 +404,80 @@ def test_parallel_transform_worker_apply() -> None:
             transform_payload_expand,
             input_batch,
             output_block,
-            index,
+            result.next_index_to_process,
         )
     smm.shutdown()
 
 
+NO_KEY = "No Key"
+
+
+def fail_bad_messages(message: Message[KafkaPayload]) -> KafkaPayload:
+    if message.payload.key is None:
+        raise InvalidMessages(
+            [
+                InvalidKafkaMessage(
+                    payload=str(message.payload),
+                    timestamp=message.timestamp,
+                    topic=message.partition.topic.name,
+                    consumer_group="",
+                    partition=message.partition.index,
+                    offset=message.offset,
+                    headers=message.payload.headers,
+                    reason=NO_KEY,
+                )
+            ]
+        )
+    return message.payload
+
+
+def test_parallel_transform_worker_bad_messages() -> None:
+    smm = SharedMemoryManager()
+    smm.start()
+    input_block = smm.SharedMemory(128)
+    output_block = smm.SharedMemory(128)
+
+    # every other message has a key
+    messages = [
+        Message(
+            Partition(Topic("test"), 0),
+            i,
+            KafkaPayload(None if i % 2 == 0 else b"key", b"\x00", []),
+            datetime.now(),
+        )
+        for i in range(9)
+    ]
+
+    input_batch = MessageBatch[Any](input_block)
+    for message in messages:
+        input_batch.append(message)
+    assert len(input_batch) == 9
+    # process entire batch
+    result = parallel_transform_worker_apply(
+        fail_bad_messages, input_batch, output_block
+    )
+    # all 9 messages processed
+    assert result.next_index_to_process == 9
+    # 5 were bad, 4 were good
+    assert len(result.invalid_messages) == 5
+    assert len(result.valid_messages_transformed) == 4
+
+    input_batch = MessageBatch[Any](input_block)
+    for message in messages:
+        input_batch.append(message)
+    assert len(input_batch) == 9
+    # process batch from halfway through
+    result = parallel_transform_worker_apply(
+        fail_bad_messages, input_batch, output_block, start_index=5
+    )
+    # all 9 messages processed
+    assert result.next_index_to_process == 9
+    # Out of remaining 4, 2 were bad, 2 were good
+    assert len(result.invalid_messages) == 2
+    assert len(result.valid_messages_transformed) == 2
+    smm.shutdown()
+
+
 def get_subprocess_count() -> int:
     return len(multiprocessing.active_children())
 
@@ -419,7 +492,7 @@ def test_parallel_transform_step() -> None:
             KafkaPayload(None, b"\x00" * size, []),
             datetime.now(),
         )
-        for i, size in enumerate([1000, 1000, 2000, 2000])
+        for i, size in enumerate([4000, 4000, 8000, 2000])
     ]
 
     starting_processes = get_subprocess_count()
@@ -439,7 +512,7 @@ def test_parallel_transform_step() -> None:
             GaugeCall("transform.processes", 2.0, tags=None),
             GaugeCall("batches_in_progress", 1.0, tags=None),
             TimingCall("batch.size.msg", 3, None),
-            TimingCall("batch.size.bytes", 4000, None),
+            TimingCall("batch.size.bytes", 16000, None),
             GaugeCall("batches_in_progress", 2.0, tags=None),
             TimingCall("batch.size.msg", 1, None),
             TimingCall("batch.size.bytes", 2000, None),
@@ -448,11 +521,11 @@ def test_parallel_transform_step() -> None:
         transform_step = ParallelTransformStep(
             transform_payload_expand,
             next_step,
-            processes=worker_processes,
+            num_processes=worker_processes,
             max_batch_size=5,
             max_batch_time=60,
-            input_block_size=4096,
-            output_block_size=4096,
+            input_block_size=16384,
+            output_block_size=16384,
         )
 
         for message in messages:
@@ -492,7 +565,7 @@ def test_parallel_transform_step_terminate_workers() -> None:
         transform_step = ParallelTransformStep(
             transform_payload_expand,  # doesn't matter
             next_step,
-            processes=worker_processes,
+            num_processes=worker_processes,
             max_batch_size=5,
             max_batch_time=60,
             input_block_size=4096,
@@ -505,3 +578,59 @@ def test_parallel_transform_step_terminate_workers() -> None:
         starting_processes,
     ), assert_changes(lambda: int(next_step.terminate.call_count), 0, 1):
         transform_step.terminate()
+
+
+def test_parallel_transform_step_bad_messages() -> None:
+    next_step = Mock()
+
+    starting_processes = get_subprocess_count()
+    worker_processes = 5
+    manager_processes = 1
+
+    # every other message has a key
+    messages = [
+        Message(
+            Partition(Topic("test"), 0),
+            0,
+            KafkaPayload(None if i % 2 == 0 else b"key", b"\x00", []),
+            datetime.now(),
+        )
+        for i in range(9)
+    ]
+
+    # everything should be processed in parallel, 5 workers should spawn
+    with assert_changes(
+        get_subprocess_count,
+        starting_processes,
+        starting_processes + worker_processes + manager_processes,
+    ):
+        # create transform step with multiple processes
+        transform_step = ParallelTransformStep(
+            function=fail_bad_messages,
+            next_step=next_step,
+            num_processes=worker_processes,
+            max_batch_size=9,
+            max_batch_time=60,
+            input_block_size=4096,
+            output_block_size=4096,
+        )
+
+        # submit 9 messages: 4 good ones 5 bad ones
+        for message in messages:
+            transform_step.submit(message)
+            transform_step.poll()
+
+    # wait for all processes to finish
+    with pytest.raises(InvalidMessages) as e_info:
+        transform_step.close()
+        transform_step.join()
+
+    # An exception should have been thrown with the 5 bad messages
+    assert len(e_info.value.messages) == 5
+    # Test exception pickles and decodes correctly
+    invalid_message = e_info.value.messages[0]
+    assert isinstance(invalid_message, InvalidKafkaMessage)
+    assert invalid_message.reason == NO_KEY
+    assert invalid_message.payload == str(messages[0].payload)
+    # The 4 good ones should not have been blocked
+    assert next_step.submit.call_count == 4
